@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2008 Pierre-Luc Beaudoin <pierre-luc@pierlux.com>
+ * Copyright (C) 2011 Jiri Techet <techet@gmail.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -100,8 +101,10 @@ struct _ChamplainLabelPrivate
   ClutterActor *shadow;
   ClutterActor *background;
   guint redraw_id;
-  
-  ClutterGroup *content_group;  
+  guint total_width;
+  guint total_height;
+
+  ClutterGroup *content_group;
 };
 
 G_DEFINE_TYPE (ChamplainLabel, champlain_label, CHAMPLAIN_TYPE_MARKER);
@@ -111,23 +114,6 @@ G_DEFINE_TYPE (ChamplainLabel, champlain_label, CHAMPLAIN_TYPE_MARKER);
 
 static void draw_label (ChamplainLabel *label);
 
-static void paint (ClutterActor *self);
-static void pick (ClutterActor *self, 
-    const ClutterColor *color);
-static void get_preferred_width (ClutterActor *self,
-    gfloat for_height,
-    gfloat *min_width_p,
-    gfloat *natural_width_p);
-static void get_preferred_height (ClutterActor *self,
-    gfloat for_width,
-    gfloat *min_height_p,
-    gfloat *natural_height_p);
-static void allocate (ClutterActor *self,
-    const ClutterActorBox *box,
-    ClutterAllocationFlags flags);
-static void map (ClutterActor *self);
-static void unmap (ClutterActor *self);
-
 
 static void
 champlain_label_get_property (GObject *object,
@@ -257,28 +243,133 @@ champlain_label_set_property (GObject *object,
 }
 
 
+
+static void
+paint (ClutterActor *self)
+{
+  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
+
+  clutter_actor_paint (CLUTTER_ACTOR (priv->content_group));
+}
+
+
+#define RADIUS 10
+#define PADDING (RADIUS / 2)
+
+static void
+pick (ClutterActor *self,
+    const ClutterColor *color)
+{
+  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
+  gfloat width, height;
+
+  if (!clutter_actor_should_pick_paint (self))
+    return;
+
+  width = priv->total_width;
+  height = priv->total_height;
+
+  cogl_path_new ();
+
+  cogl_set_source_color4ub (color->red,
+      color->green,
+      color->blue,
+      color->alpha);
+
+  cogl_path_move_to (RADIUS, 0);
+  cogl_path_line_to (width - RADIUS, 0);
+  cogl_path_arc (width - RADIUS, RADIUS, RADIUS, RADIUS, -90, 0);
+  cogl_path_line_to (width, height - RADIUS);
+  cogl_path_arc (width - RADIUS, height - RADIUS, RADIUS, RADIUS, 0, 90);
+  cogl_path_line_to (RADIUS, height);
+  cogl_path_arc (RADIUS, height - RADIUS, RADIUS, RADIUS, 90, 180);
+  cogl_path_line_to (0, RADIUS);
+  cogl_path_arc (RADIUS, RADIUS, RADIUS, RADIUS, 180, 270);
+  cogl_path_close ();
+  cogl_path_fill ();
+}
+
+
+static void
+get_preferred_width (ClutterActor *self,
+    gfloat for_height,
+    gfloat *min_width_p,
+    gfloat *natural_width_p)
+{
+  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
+
+  clutter_actor_get_preferred_width (CLUTTER_ACTOR (priv->content_group),
+      for_height,
+      min_width_p,
+      natural_width_p);
+}
+
+
+static void
+get_preferred_height (ClutterActor *self,
+    gfloat for_width,
+    gfloat *min_height_p,
+    gfloat *natural_height_p)
+{
+  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
+
+  clutter_actor_get_preferred_height (CLUTTER_ACTOR (priv->content_group),
+      for_width,
+      min_height_p,
+      natural_height_p);
+}
+
+
+static void
+allocate (ClutterActor *self,
+    const ClutterActorBox *box,
+    ClutterAllocationFlags flags)
+{
+  ClutterActorBox child_box;
+
+  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
+
+  CLUTTER_ACTOR_CLASS (champlain_label_parent_class)->allocate (self, box, flags);
+
+  child_box.x1 = 0;
+  child_box.x2 = box->x2 - box->x1;
+  child_box.y1 = 0;
+  child_box.y2 = box->y2 - box->y1;
+
+  clutter_actor_allocate (CLUTTER_ACTOR (priv->content_group), &child_box, flags);
+}
+
+
+static void
+map (ClutterActor *self)
+{
+  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
+
+  CLUTTER_ACTOR_CLASS (champlain_label_parent_class)->map (self);
+
+  clutter_actor_map (CLUTTER_ACTOR (priv->content_group));
+}
+
+
+static void
+unmap (ClutterActor *self)
+{
+  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
+
+  CLUTTER_ACTOR_CLASS (champlain_label_parent_class)->unmap (self);
+
+  clutter_actor_unmap (CLUTTER_ACTOR (priv->content_group));
+}
+
+
 static void
 champlain_label_dispose (GObject *object)
 {
   ChamplainLabelPrivate *priv = CHAMPLAIN_LABEL (object)->priv;
 
-  if (priv->background)
-    {
-      g_object_unref (priv->background);
-      priv->background = NULL;
-    }
-
-  if (priv->shadow)
-    {
-      g_object_unref (priv->shadow);
-      priv->shadow = NULL;
-    }
-
-  if (priv->text_actor)
-    {
-      g_object_unref (priv->text_actor);
-      priv->text_actor = NULL;
-    }
+  priv->background = NULL;
+  priv->shadow = NULL;
+  priv->text_actor = NULL;
 
   if (priv->image)
     {
@@ -346,7 +437,7 @@ champlain_label_class_init (ChamplainLabelClass *klass)
 {
   ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
-  
+
   g_type_class_add_private (klass, sizeof (ChamplainLabelPrivate));
 
   object_class->finalize = champlain_label_finalize;
@@ -370,8 +461,11 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_TEXT,
-      g_param_spec_string ("text", "Text", "The text of the label",
-          "", CHAMPLAIN_PARAM_READWRITE));
+      g_param_spec_string ("text", 
+          "Text", 
+          "The text of the label",
+          "", 
+          CHAMPLAIN_PARAM_READWRITE));
 
   /**
    * ChamplainLabel:image:
@@ -381,8 +475,11 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_IMAGE,
-      g_param_spec_object ("image", "Image", "The image of the label",
-          CLUTTER_TYPE_ACTOR, CHAMPLAIN_PARAM_READWRITE));
+      g_param_spec_object ("image", 
+          "Image", 
+          "The image of the label",
+          CLUTTER_TYPE_ACTOR, 
+          CHAMPLAIN_PARAM_READWRITE));
 
   /**
    * ChamplainLabel:use-markup:
@@ -392,8 +489,11 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_USE_MARKUP,
-      g_param_spec_boolean ("use-markup", "Use Markup", "The text uses markup",
-          FALSE, CHAMPLAIN_PARAM_READWRITE));
+      g_param_spec_boolean ("use-markup", 
+          "Use Markup", 
+          "The text uses markup",
+          FALSE, 
+          CHAMPLAIN_PARAM_READWRITE));
 
   /**
    * ChamplainLabel:alignment:
@@ -403,8 +503,12 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_ALIGNMENT,
-      g_param_spec_enum ("alignment", "Alignment", "The label's alignment",
-          PANGO_TYPE_ALIGNMENT, PANGO_ALIGN_LEFT, CHAMPLAIN_PARAM_READWRITE));
+      g_param_spec_enum ("alignment", 
+          "Alignment", 
+          "The label's alignment",
+          PANGO_TYPE_ALIGNMENT, 
+          PANGO_ALIGN_LEFT, 
+          CHAMPLAIN_PARAM_READWRITE));
 
   /**
    * ChamplainLabel:color:
@@ -414,8 +518,11 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_COLOR,
-      clutter_param_spec_color ("color", "Color", "The label's color",
-          &DEFAULT_COLOR, CHAMPLAIN_PARAM_READWRITE));
+      clutter_param_spec_color ("color", 
+          "Color", 
+          "The label's color",
+          &DEFAULT_COLOR, 
+          CHAMPLAIN_PARAM_READWRITE));
 
   /**
    * ChamplainLabel:text-color:
@@ -425,8 +532,11 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_TEXT_COLOR,
-      clutter_param_spec_color ("text-color", "Text Color", "The label's text color",
-          &DEFAULT_TEXT_COLOR, CHAMPLAIN_PARAM_READWRITE));
+      clutter_param_spec_color ("text-color", 
+          "Text Color", 
+          "The label's text color",
+          &DEFAULT_TEXT_COLOR, 
+          CHAMPLAIN_PARAM_READWRITE));
 
   /**
    * ChamplainLabel:font-name:
@@ -436,8 +546,11 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_FONT_NAME,
-      g_param_spec_string ("font-name", "Font Name", "The label's text font name",
-          "Sans 11", CHAMPLAIN_PARAM_READWRITE));
+      g_param_spec_string ("font-name", 
+          "Font Name", 
+          "The label's text font name",
+          "Sans 11", 
+          CHAMPLAIN_PARAM_READWRITE));
 
   /**
    * ChamplainLabel:wrap:
@@ -447,8 +560,11 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_WRAP,
-      g_param_spec_boolean ("wrap", "Wrap", "The label's text wrap",
-          FALSE, CHAMPLAIN_PARAM_READWRITE));
+      g_param_spec_boolean ("wrap", 
+          "Wrap", 
+          "The label's text wrap",
+          FALSE, 
+          CHAMPLAIN_PARAM_READWRITE));
 
   /**
    * ChamplainLabel:wrap-mode:
@@ -458,8 +574,12 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_WRAP_MODE,
-      g_param_spec_enum ("wrap-mode", "Wrap Mode", "The label's text wrap mode",
-          PANGO_TYPE_WRAP_MODE, PANGO_WRAP_WORD, CHAMPLAIN_PARAM_READWRITE));
+      g_param_spec_enum ("wrap-mode", 
+          "Wrap Mode", 
+          "The label's text wrap mode",
+          PANGO_TYPE_WRAP_MODE, 
+          PANGO_WRAP_WORD, 
+          CHAMPLAIN_PARAM_READWRITE));
 
   /**
    * ChamplainLabel:ellipsize:
@@ -469,8 +589,12 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_ELLIPSIZE,
-      g_param_spec_enum ("ellipsize", "Ellipsize Mode", "The label's text ellipsize mode",
-          PANGO_TYPE_ELLIPSIZE_MODE, PANGO_ELLIPSIZE_NONE, CHAMPLAIN_PARAM_READWRITE));
+      g_param_spec_enum ("ellipsize", 
+          "Ellipsize Mode", 
+          "The label's text ellipsize mode",
+          PANGO_TYPE_ELLIPSIZE_MODE, 
+          PANGO_ELLIPSIZE_NONE, 
+          CHAMPLAIN_PARAM_READWRITE));
 
   /**
    * ChamplainLabel:draw-background:
@@ -480,8 +604,11 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_DRAW_BACKGROUND,
-      g_param_spec_boolean ("draw-background", "Draw Background", "The label has a background",
-          TRUE, CHAMPLAIN_PARAM_READWRITE));
+      g_param_spec_boolean ("draw-background", 
+          "Draw Background", 
+          "The label has a background",
+          TRUE, 
+          CHAMPLAIN_PARAM_READWRITE));
 
   /**
    * ChamplainLabel:single-line-mode:
@@ -491,14 +618,14 @@ champlain_label_class_init (ChamplainLabelClass *klass)
    * Since: 0.10
    */
   g_object_class_install_property (object_class, PROP_SINGLE_LINE_MODE,
-      g_param_spec_boolean ("single-line-mode", "Single Line Mode", "The label's single line mode",
-          TRUE, CHAMPLAIN_PARAM_READWRITE));
+      g_param_spec_boolean ("single-line-mode", 
+          "Single Line Mode", 
+          "The label's single line mode",
+          TRUE, 
+          CHAMPLAIN_PARAM_READWRITE));
 }
 
 
-#define RADIUS 10
-#define PADDING (RADIUS / 2)
-
 static void
 draw_box (cairo_t *cr,
     gint width,
@@ -559,7 +686,7 @@ draw_shadow (ChamplainLabel *label,
   cr = clutter_cairo_texture_create (CLUTTER_CAIRO_TEXTURE (shadow));
 
   cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
-  cairo_paint(cr);
+  cairo_paint (cr);
   cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
 
   cairo_matrix_init (&matrix,
@@ -583,10 +710,9 @@ draw_shadow (ChamplainLabel *label,
     {
       clutter_container_remove_actor (CLUTTER_CONTAINER (priv->content_group),
           priv->shadow);
-      g_object_unref (priv->shadow);
     }
 
-  priv->shadow = g_object_ref (shadow);
+  priv->shadow = shadow;
 }
 
 
@@ -607,7 +733,7 @@ draw_background (ChamplainLabel *label,
   cr = clutter_cairo_texture_create (CLUTTER_CAIRO_TEXTURE (bg));
 
   cairo_set_operator (cr, CAIRO_OPERATOR_CLEAR);
-  cairo_paint(cr);
+  cairo_paint (cr);
   cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
 
   /* If selected, add the selection color to the marker's color */
@@ -643,10 +769,9 @@ draw_background (ChamplainLabel *label,
     {
       clutter_container_remove_actor (CLUTTER_CONTAINER (priv->content_group),
           priv->background);
-      g_object_unref (priv->background);
     }
 
-  priv->background = g_object_ref (bg);
+  priv->background = bg;
 }
 
 
@@ -673,7 +798,6 @@ draw_label (ChamplainLabel *label)
       if (priv->text_actor == NULL)
         {
           priv->text_actor = clutter_text_new_with_text (priv->font_name, priv->text);
-          g_object_ref (priv->text_actor);
         }
 
       text = CLUTTER_TEXT (priv->text_actor);
@@ -715,6 +839,8 @@ draw_label (ChamplainLabel *label)
     }
 
   point = (total_height + 2 * PADDING) / 4.0;
+  priv->total_width = total_width;
+  priv->total_height = total_height;
 
   if (priv->draw_background)
     {
@@ -726,14 +852,12 @@ draw_label (ChamplainLabel *label)
       if (priv->background != NULL)
         {
           clutter_container_remove_actor (CLUTTER_CONTAINER (priv->content_group), priv->background);
-          g_object_unref (G_OBJECT (priv->background));
           priv->background = NULL;
         }
 
       if (priv->shadow != NULL)
         {
           clutter_container_remove_actor (CLUTTER_CONTAINER (priv->content_group), priv->shadow);
-          g_object_unref (G_OBJECT (priv->shadow));
           priv->shadow = NULL;
         }
     }
@@ -758,8 +882,6 @@ draw_label (ChamplainLabel *label)
     clutter_actor_set_anchor_point (CLUTTER_ACTOR (label),
         0,
         clutter_actor_get_height (priv->text_actor) / 2.0);
-
-  clutter_actor_queue_redraw (CLUTTER_ACTOR (label));
 }
 
 
@@ -837,6 +959,8 @@ champlain_label_init (ChamplainLabel *label)
   priv->content_group = CLUTTER_GROUP (clutter_group_new ());
   clutter_actor_set_parent (CLUTTER_ACTOR (priv->content_group), CLUTTER_ACTOR (label));
   clutter_actor_queue_relayout (CLUTTER_ACTOR (label));
+  priv->total_width = 0;
+  priv->total_height = 0;
 
   g_signal_connect (label, "notify::selected", G_CALLBACK (notify_selected), NULL);
 }
@@ -858,99 +982,6 @@ champlain_label_new (void)
 }
 
 
-static void
-paint (ClutterActor *self)
-{
-  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
-  
-  clutter_actor_paint (CLUTTER_ACTOR (priv->content_group));
-}
-
-
-static void
-pick (ClutterActor *self, 
-    const ClutterColor *color)
-{
-  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
-
-  CLUTTER_ACTOR_CLASS (champlain_label_parent_class)->pick (self, color);
-
-  clutter_actor_paint (CLUTTER_ACTOR (priv->content_group));
-}
-
-
-static void
-get_preferred_width (ClutterActor *self,
-    gfloat for_height,
-    gfloat *min_width_p,
-    gfloat *natural_width_p)
-{
-  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
-
-  clutter_actor_get_preferred_width (CLUTTER_ACTOR (priv->content_group),
-      for_height,
-      min_width_p,
-      natural_width_p);
-}
-
-
-static void
-get_preferred_height (ClutterActor *self,
-    gfloat for_width,
-    gfloat *min_height_p,
-    gfloat *natural_height_p)
-{
-  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
-
-  clutter_actor_get_preferred_height (CLUTTER_ACTOR (priv->content_group),
-      for_width,
-      min_height_p,
-      natural_height_p);
-}
-
-
-static void
-allocate (ClutterActor *self,
-    const ClutterActorBox *box,
-    ClutterAllocationFlags flags)
-{
-  ClutterActorBox child_box;
-
-  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
-
-  CLUTTER_ACTOR_CLASS (champlain_label_parent_class)->allocate (self, box, flags);
-
-  child_box.x1 = 0;
-  child_box.x2 = box->x2 - box->x1;
-  child_box.y1 = 0;
-  child_box.y2 = box->y2 - box->y1;
-
-  clutter_actor_allocate (CLUTTER_ACTOR (priv->content_group), &child_box, flags);
-}
-
-
-static void
-map (ClutterActor *self)
-{
-  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
-
-  CLUTTER_ACTOR_CLASS (champlain_label_parent_class)->map (self);
-
-  clutter_actor_map (CLUTTER_ACTOR (priv->content_group));
-}
-
-
-static void
-unmap (ClutterActor *self)
-{
-  ChamplainLabelPrivate *priv = GET_PRIVATE (self);
-
-  CLUTTER_ACTOR_CLASS (champlain_label_parent_class)->unmap (self);
-
-  clutter_actor_unmap (CLUTTER_ACTOR (priv->content_group));
-}
-
-
 /**
  * champlain_label_new_with_text:
  * @text: the text of the text
@@ -1004,9 +1035,7 @@ champlain_label_new_with_image (ClutterActor *actor)
   ChamplainLabel *label = CHAMPLAIN_LABEL (champlain_label_new ());
 
   if (actor != NULL)
-    {
-      champlain_label_set_image (label, actor);
-    }
+    champlain_label_set_image (label, actor);
 
   return CLUTTER_ACTOR (label);
 }
@@ -1092,6 +1121,7 @@ champlain_label_set_text (ChamplainLabel *label,
     g_free (priv->text);
 
   priv->text = g_strdup (text);
+  g_object_notify (G_OBJECT (label), "text");
   champlain_label_queue_redraw (label);
 }
 
@@ -1298,7 +1328,7 @@ champlain_label_set_wrap_mode (ChamplainLabel *label,
   g_return_if_fail (CHAMPLAIN_IS_LABEL (label));
 
   label->priv->wrap_mode = wrap_mode;
-  g_object_notify (G_OBJECT (label), "wrap");
+  g_object_notify (G_OBJECT (label), "wrap-mode");
   champlain_label_queue_redraw (label);
 }
 
@@ -1623,3 +1653,22 @@ champlain_label_get_draw_background (ChamplainLabel *label)
 
   return label->priv->draw_background;
 }
+
+
+/**
+ * champlain_label_get_attributes:
+ * @label: The label
+ *
+ * Gets the label's text attributes.
+ *
+ * Returns: the label's text attributes.
+ *
+ * Since: 0.10
+ */
+PangoAttrList *
+champlain_label_get_attributes (ChamplainLabel *label)
+{
+  g_return_val_if_fail (CHAMPLAIN_IS_LABEL (label), NULL);
+
+  return label->priv->attributes;
+}
